/*
* Copyright (c) 2011-2020 LightPad
*               2021 Zorin OS Technologies Ltd.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public
* License as published by the Free Software Foundation; either
* version 2 of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* General Public License for more details.
*
* You should have received a copy of the GNU General Public
* License along with this program; if not, write to the
* Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
* Boston, MA 02110-1301 USA
*/

// We need to probe the dbus daemon directly, hence this interface
[DBus (name="org.freedesktop.DBus")]
public interface DBusImpl : Object {
    public abstract async string[] list_names() throws DBusError, IOError;
    public signal void name_owner_changed(string name, string old_owner, string new_owner);
}

public class ZorinAppGridLiteWindow : Widgets.CompositedWindow {

    public static string user_home = GLib.Environment.get_variable ("HOME");
    public Gee.ArrayList<Gee.HashMap<string, string>> apps = new Gee.ArrayList<Gee.HashMap<string, string>> ();
    public Gee.HashMap<string, Gdk.Pixbuf> icons = new Gee.HashMap<string, Gdk.Pixbuf>();
    public Gee.ArrayList<Gee.HashMap<string, string>> filtered = new Gee.ArrayList<Gee.HashMap<string, string>> ();
    public ZorinAppGridLite.Frontend.Indicators pages;

    public double font_size;
    public int icon_size;
    public int item_box_width;
    public int item_box_height;

    public int total_pages;
    public int scroll_times = 0;
    public int SCROLL_SENSITIVITY = 12;

    public Gdk.Rectangle monitor_dimensions;
    public Gtk.Box top_spacer;
    public GLib.List<ZorinAppGridLite.Frontend.AppItemContainer> children = new GLib.List<ZorinAppGridLite.Frontend.AppItemContainer> ();
    public Gtk.SearchEntry search_entry;
    public Gtk.Grid grid;

    private int grid_x;
    private int grid_y;

    public bool dynamic_background = false;
    public double factor_scaling;
    public string file_png = user_home + Resources.ZORIN_APPGRID_LITE_CONFIG_DIR +
        "/" + "background.png";
    public string file_jpg = user_home + Resources.ZORIN_APPGRID_LITE_CONFIG_DIR +
        "/" + "background.jpg";
    public Cairo.Pattern pattern;
    public Cairo.ImageSurface image_sf;
    public Gdk.Pixbuf image_pf;
    
    public bool focus_quit = true;
    
    public DBusImpl? impl = null;
    public string wanted_dbus_id = "";
    public HashTable<string,bool> active_dbus_names = null;
    

    public ZorinAppGridLiteWindow () {
        // There isn't always a primary monitor.
        Gdk.Monitor monitor = get_display ().get_primary_monitor () ?? get_display ().get_monitor (0);
        Gdk.Rectangle pixel_geo = monitor.get_geometry ();
        // get_geometry() returns "device pixels", but we need "application pixels".
        monitor_dimensions.width = pixel_geo.width / monitor.get_scale_factor ();
        monitor_dimensions.height = pixel_geo.height / monitor.get_scale_factor ();

        FileConfig config = new FileConfig(
            monitor_dimensions.width, 
            monitor_dimensions.height, 
            user_home + Resources.CONFIG_FILE
        );
        // For compatibility, maybe add FileConfig to ZorinAppGridLiteWindow someday
        this.icon_size = config.item_icon_size;
        this.font_size = config.item_font_size;
        this.item_box_width = config.item_box_width;
        this.item_box_height = config.item_box_height;
        this.grid_y = config.grid_y;
        this.grid_x = config.grid_x;

        message ("The monitor dimensions are: %dx%d", monitor_dimensions.width,  monitor_dimensions.height);
        message ("The apps icon size is: %d", this.icon_size);
        message ("The grid size are: %dx%d", this.grid_y, this.grid_x);

        // Window properties
        this.set_title ("Zorin AppGrid Lite");
        /* Skip that a workspace switcher and taskbars displays a
           thumbnail representation of the window in the screen */
        this.set_skip_pager_hint (true);
        this.set_skip_taskbar_hint (true);
        this.set_type_hint (Gdk.WindowTypeHint.NORMAL);
        this.move(pixel_geo.x, pixel_geo.y);
        this.fullscreen ();
        this.set_default_size (monitor_dimensions.width,  monitor_dimensions.height);

        // Get all apps
        ZorinAppGridLite.Backend.DesktopEntries.enumerate_apps (this.icons, this.icon_size, user_home, out this.apps);

        // Add container wrapper
        var wrapper = new Gtk.EventBox (); // Used for the scrolling and button press events
        wrapper.set_visible_window (false);
        this.add (wrapper);

        // Add container
        var container = new Gtk.Box (Gtk.Orientation.VERTICAL, 10);
        wrapper.add (container);

        // Add top bar
        var bottom = new Gtk.Box (Gtk.Orientation.HORIZONTAL, 0);

        // Search Entry
        this.search_entry = new Gtk.SearchEntry ();
        this.search_entry.placeholder_text =  _("Type to search");
        this.search_entry.set_size_request (240, 32);
        message ("Search Entry created");

        // Lateral distance (120 are the pixels of the search entry width)
        int screen_half = (monitor_dimensions.width / 2) - 120;
        bottom.pack_start (this.search_entry, false, true, screen_half);

        // Upstairs (padding is the space between search entry and the grid)
        container.pack_start (bottom, false, true, 32);

        this.grid = new Gtk.Grid();
        this.grid.row_homogeneous = true;
        this.grid.column_homogeneous = true;
        this.grid.set_row_spacing (config.grid_row_spacing);
        this.grid.set_column_spacing (config.grid_col_spacing);
        this.grid.set_halign (Gtk.Align.CENTER);
        this.grid.set_valign (Gtk.Align.CENTER);

        // Initialize the grid
        for (int c = 0; c < this.grid_y; c++) {
            this.grid.insert_column (c);
        }

        for (int r = 0; r < this.grid_x; r++) {
            this.grid.insert_row (r);
        }

        container.pack_start (this.grid, true, true, 0);

        this.populate_grid ();

        // Add pages
        this.pages = new ZorinAppGridLite.Frontend.Indicators ();
        this.pages.child_activated.connect ( () => { this.update_grid (this.filtered); } );

        var pages_wrapper = new Gtk.Box (Gtk.Orientation.HORIZONTAL, 0);
        pages_wrapper.set_size_request (-1, 30);
        container.pack_end (pages_wrapper, false, true, 15);

        // Find number of pages and populate
        // First order the apps alphabetically
        this.apps.sort ((a, b) => GLib.strcmp (a["name"], b["name"]));
        this.update_pages (this.apps);
        if (this.total_pages > 1) {
            pages_wrapper.pack_start (this.pages, true, false, 0);
            for (int p = 1; p <= this.total_pages; p++) {
                // Add the number of pages as text
                //this.pages.append (p.to_string ());
                this.pages.append ("⬤");
            }
        }
        this.pages.set_active (0);

        // Signals and callbacks
        this.search_entry.key_press_event.connect (on_search_entry_key_press);
        this.search_entry.button_release_event.connect ( () => { return Gdk.EVENT_STOP; });
        wrapper.key_press_event.connect (on_wrapper_key_press);
        this.add_events (Gdk.EventMask.SCROLL_MASK);
        this.button_release_event.connect ( () => { this.destroy (); return Gdk.EVENT_STOP; });
        // Dynamic Background
        if (GLib.File.new_for_path (file_jpg).query_exists ()) {
            this.dynamic_background = true;
            try {
                image_pf = new Gdk.Pixbuf.from_file (file_jpg);
                int w = image_pf.get_width ();
                factor_scaling = (double) ((double) ((monitor_dimensions.width * 100) / w) / 100);
            } catch (GLib.Error e) {
                warning ("Cant create Pixbuf background!");
            }
        } else if (GLib.File.new_for_path (file_png).query_exists ()) {
            this.dynamic_background = true;
            image_sf = new Cairo.ImageSurface.from_png (file_png);
            pattern = new Cairo.Pattern.for_surface (image_sf);
            pattern.set_extend (Cairo.Extend.PAD);
            int w = image_sf.get_width ();
            factor_scaling = (double) ((double) ((monitor_dimensions.width * 100) / w) / 100);
        }

        this.draw.connect (this.draw_background);
        /* Disabling because it causes the menu to re-open when the keyboard shortcut is invoked.
        this.focus_out_event.connect ( () => {
            if (this.focus_quit) {
                this.destroy ();
            }
            return Gdk.EVENT_STOP;
        } );
        */
        this.search_entry.search_changed.connect (this.search);
        this.search_entry.text = "";
        
        // Setup dbus monitoring
        active_dbus_names = new HashTable<string,bool>(str_hash, str_equal);
        setup_dbus.begin();
        
        this.search();
    }

    private void search() {
        var current_text = this.search_entry.text.strip ().down ();
        this.filtered.clear ();

        foreach (Gee.HashMap<string, string> app in this.apps) {
            if ((app["name"] != null && current_text in app["name"].down ()) ||
                (app["description"] != null && current_text in app["description"].down ()) ||
                (app["command"] != null && current_text in app["command"].down ())) {
                this.filtered.add (app);
            }
        }

        this.pages.set_active (0);

        this.queue_draw ();
    }

    private void populate_grid () {
        for (int r = 0; r < this.grid_x; r++) {
            for (int c = 0; c < this.grid_y; c++) {
                var item = new ZorinAppGridLite.Frontend.AppItem (
                    this.icon_size, this.font_size, 
                    this.item_box_width, this.item_box_height
                );
                var item_container = new ZorinAppGridLite.Frontend.AppItemContainer(item);
                this.children.append (item_container);

                item.enter_notify_event.connect ( () => { item.grab_focus (); return Gdk.EVENT_STOP; } );
                item.button_release_event.connect ( () => {
                    try {
                    	this.focus_quit = false;
                        int child_index = this.children.index (item_container);
                        int page_active = this.pages.active;
                        /* Prevent indicators pages to get a negative one (-1)
                           and fix with this the bug 003 where a negative result
                           is obtained and that index does not exist */
                        if (page_active < 0) {
                            page_active = 0;
                        }
                        int app_index = (int) (child_index + (page_active * this.grid_y * this.grid_x));

                        /* GTK+ implements open apps in terminal in this way:
                         * https://github.com/GNOME/glib/blob/cd1eba043c90da3aee8f5cd51b205b2e2c16f08e/gio/gdesktopappinfo.c#L2467-L2494
                         * So, if the desktop environment is not GNOME we need xterm as dependency
                         * and the way to open apps in terminal is with the following code:
                         */
                        if (this.filtered.get (app_index)["terminal"] == "true") {
                            GLib.AppInfo.create_from_commandline (this.filtered.get (app_index)["command"], null, GLib.AppInfoCreateFlags.NEEDS_TERMINAL).launch (null, null);
                        } else {
                            new GLib.DesktopAppInfo.from_filename (this.filtered.get(app_index)["desktop_file"]).launch (null, null);
                        }
                        if (this.filtered.get (app_index)["dbus"] == "true") {
                            this.wanted_dbus_id = this.filtered.get (app_index)["dbus_id"];
                            this.check_dbus_name();
                            /* Some apps are slow to open so hide and quit when they're done */
                            this.hide();
                        } else {
                            this.destroy ();
                        }
                    } catch (GLib.Error e) {
                        warning ("Error! Load application: " + e.message);
                    }

                    return Gdk.EVENT_STOP;
                });
                this.grid.attach (item_container, c, r, 1, 1);
            }
        }
    }

    private void update_grid (Gee.ArrayList<Gee.HashMap<string, string>> apps) {
        int item_iter = (int)(this.pages.active * this.grid_y * this.grid_x);
        /* Fix for bug 001 with message:
        arraylist.c:1181:gee_array_list_real_get: assertion failed: (index >= 0) */
        if (item_iter < 0) {
            item_iter = 0;
        }

        for (int r = 0; r < this.grid_x; r++) {
            for (int c = 0; c < this.grid_y; c++) {
                int table_pos = c + (r * (int)this.grid_y); // position in table right now

                var item = this.children.nth_data (table_pos);
                if (item_iter < apps.size) {
                    var current_item = apps.get(item_iter);

                    // Update app
                    if (current_item["description"] == null || current_item["description"] == "") {
                        item.change_app (icons[current_item["command"]], current_item["name"], current_item["name"]);
                    } else {
                        item.change_app (icons[current_item["command"]], current_item["name"], current_item["name"] + ":\n" + current_item["description"]);
                    }
                    item.show_app_item();
                } else { // fill with a blank one
                    item.hide_app_item();
                }

                item_iter++;

            }
        }

        // Update number of pages
        this.update_pages (apps);

        // Grab first one's focus
        if (this.children.nth_data (0).is_item_visible()) {
            this.children.nth_data (0).grab_focus ();
        }
    }

    private void update_pages (Gee.ArrayList<Gee.HashMap<string, string>> apps) {
        // Find current number of pages and update count
        var num_pages = (int) (apps.size / (this.grid_y * this.grid_x));
        (double) apps.size % (double) (this.grid_y * this.grid_x) > 0 ? this.total_pages = num_pages + 1 : this.total_pages = num_pages;

        // Update pages
        if (this.total_pages > 1) {
            this.pages.visible = true;
            for (int p = 1; p <= this.pages.children.length (); p++) {
                p > this.total_pages ? this.pages.children.nth_data (p - 1).visible = false : this.pages.children.nth_data (p - 1).visible = true;
            }
        } else {
            this.pages.visible = false;
        }
    }

    private void page_left () {
        if (this.pages.active >= 1) {
            this.pages.set_active (this.pages.active - 1);
        }
    }

    private void page_right () {
        if ((this.pages.active + 1) < this.total_pages) {
            this.pages.set_active (this.pages.active + 1);
        }
    }

    private bool draw_background (Gtk.Widget widget, Cairo.Context context) {
        if (this.dynamic_background) {
            if (image_pf != null) { // If JPG exist, prefer this
                context.scale (factor_scaling, factor_scaling);
                Gdk.cairo_set_source_pixbuf (context, image_pf, 0, 0);
            } else { // Is PNG image
                context.scale (factor_scaling, factor_scaling);
                context.set_source (pattern);
            }

            context.paint ();
        } else {
            // Semi-dark background
            Gtk.Allocation size;
            widget.get_allocation (out size);

            var linear_gradient = new Cairo.Pattern.linear (size.x, size.y, size.x, size.y + size.height);
            linear_gradient.add_color_stop_rgba (0.0, 0.0, 0.0, 0.0, 0.85);

            context.set_source (linear_gradient);
            context.paint ();
        }

        return false;
    }

    public bool on_search_entry_key_press (Gdk.EventKey event) {
        var key = Gdk.keyval_name (event.keyval).replace ("KP_", "");

        switch (key) {
            case "Down":
                search_entry.move_focus (Gtk.DirectionType.TAB_FORWARD);
                return Gdk.EVENT_STOP;

            case "Escape":
                if (search_entry.text.length > 0) {
                    search_entry.text = "";
                } else {
                    this.destroy ();
                }
                return Gdk.EVENT_STOP;
        }

        return Gdk.EVENT_PROPAGATE;
    }

    // Keyboard shortcuts
    public bool on_wrapper_key_press (Gdk.EventKey event) {
        var key = Gdk.keyval_name (event.keyval).replace ("KP_", "");


        if (((event.state & Gdk.ModifierType.MOD1_MASK) != 0) && key == "F4") {
        this.destroy();
        return Gdk.EVENT_STOP;
    }
        switch (key) {
            case "Escape":
                this.destroy ();
                return true;
            case "ISO_Left_Tab":
                this.page_left ();
                break;
            case "Tab":
                this.page_right ();
                break;
            case "Left":
                if (this.get_focus () != this.search_entry) {
                    var current_item = this.grid.get_children ().index (this.get_focus ().parent);
                    if (current_item % this.grid_y == this.grid_y - 1) {
                        this.page_left ();
                        break;
                    }
                }
                return Gdk.EVENT_PROPAGATE;
            case "Right":
                if (this.get_focus () != this.search_entry) {
                    var current_item = this.grid.get_children ().index (this.get_focus ().parent);
                    if (current_item % this.grid_y == 0) {
                        this.page_right ();
                        break;
                    }
                }
                return Gdk.EVENT_PROPAGATE;
            case "Enter": // "KP_Enter"
            case "KP_Enter":
            case "Return":
                if (this.filtered.size >= 1 && (this.get_focus () != this.search_entry)) {
                    this.get_focus ().button_release_event ((Gdk.EventButton) new Gdk.Event (Gdk.EventType.BUTTON_PRESS));
                }
                break;
            case "Down":
            case "Home":
                return Gdk.EVENT_PROPAGATE;
            case "Up":
                if (this.get_focus () != this.search_entry) {
                    var current_item = this.grid.get_children ().index (this.get_focus ().parent);
                    var current_item_col = current_item % this.grid_y;
                    var current_item_row = (current_item - current_item_col) / this.grid_y;
                    if (current_item_row % this.grid_x == this.grid_x - 1) {
                        break;
                    }
                }
                return Gdk.EVENT_PROPAGATE;
            case "BackSpace":
                if (!this.search_entry.has_focus && this.search_entry.text.length > 0) {
                    this.search_entry.grab_focus ();
                    this.search_entry.move_cursor (Gtk.MovementStep.BUFFER_ENDS, 0, false);
                    this.search_entry.key_press_event (event);
                }
                return Gdk.EVENT_PROPAGATE;
            default:
                if (!this.search_entry.has_focus && event.is_modifier != 1) {
                    this.search_entry.grab_focus ();
                    this.search_entry.move_cursor (Gtk.MovementStep.BUFFER_ENDS, 0, false);
                    this.search_entry.key_press_event (event);
                }
                return Gdk.EVENT_PROPAGATE;

        }

        return Gdk.EVENT_STOP;
    }

    // Scrolling left/right for pages
    public override bool scroll_event (Gdk.EventScroll event) {
        scroll_times += 1;
        var direction = event.direction.to_string ();
        if ((direction == "GDK_SCROLL_UP" || direction == "GDK_SCROLL_LEFT")
                                    && (scroll_times >= SCROLL_SENSITIVITY)) {
            this.page_left ();
        } else if ((direction == "GDK_SCROLL_DOWN" || direction == "GDK_SCROLL_RIGHT")
                                    && (scroll_times >= SCROLL_SENSITIVITY)) {
            this.page_right ();
        }
        // If the direction is GDK_SCROLL_SMOOTH skip it

        return false;
    }
    
    private void on_name_owner_changed(string? n, string? o, string? ne) {
        if (o == "") {
            this.active_dbus_names[n] = true;
            this.check_dbus_name();
        } else {
            if (n in this.active_dbus_names) {
                this.active_dbus_names.remove(n);
            }
        }
    }

    // Check if our dbus name appeared. if it did, quit Zorin AppGrid Lite.
    private void check_dbus_name() {
        if (this.wanted_dbus_id != "" && this.wanted_dbus_id in this.active_dbus_names) {
            this.destroy ();
        }
    }

    // Do basic dbus initialisation
    public async void setup_dbus() {
        try {
            impl = yield Bus.get_proxy(BusType.SESSION, "org.freedesktop.DBus", "/org/freedesktop/DBus");

            /* Cache the names already active */
            foreach (string name in yield impl.list_names()) {
                this.active_dbus_names[name] = true;
            }
            /* Watch for new names */
            impl.name_owner_changed.connect(on_name_owner_changed);
        } catch (Error e) {
            warning("Failed to initialise dbus: %s", e.message);
        }
    }

    // Override destroy for fade out and stuff
    public new void destroy () {
        base.destroy();
        Gtk.main_quit();
    }

}

static int main (string[] args) {

    Gtk.init (ref args);
    Gtk.Application app = new Gtk.Application ("com.zorin.appgrid-lite", GLib.ApplicationFlags.FLAGS_NONE);
    
    Intl.bindtextdomain (GETTEXT_PACKAGE, LOCALEDIR);
    Intl.bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");

    // CSS Style Provider
    // Path where takes the CSS file
    string css_file = Config.PACKAGE_SHAREDIR +
        "/" + Config.PROJECT_NAME +
        "/" + "application.css";
    var css_provider = new Gtk.CssProvider ();

    try {
        css_provider.load_from_path (css_file);
        Gtk.StyleContext.add_provider_for_screen (Gdk.Screen.get_default(), css_provider,
                                                        Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION);
    } catch (GLib.Error e) {
        warning ("Could not load CSS file: %s",css_file);
    }

    app.activate.connect( () => {
        if (app.get_windows ().length () == 0) {
            var main_window = new ZorinAppGridLiteWindow ();
            main_window.set_application (app);
            main_window.show_all ();
            Gtk.main ();
        } else {
            foreach (Gtk.Window existing_window in app.get_windows()) {
                existing_window.close();
            }
            Gtk.main_quit();
        }
    });
    app.run (args);
    return 1;

}
