/*
* Copyright (c) 2011-2020 LightPad
*               2021 Zorin OS Technologies Ltd.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public
* License as published by the Free Software Foundation; either
* version 2 of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* General Public License for more details.
*
* You should have received a copy of the GNU General Public
* License along with this program; if not, write to the
* Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
* Boston, MA 02110-1301 USA
*/

namespace ZorinAppGridLite.Frontend {

    public class AppItem : Gtk.Button {
    
        private Gdk.Pixbuf icon;
        private ZorinAppGridLite.Frontend.Color prominent;
        private Gtk.Box wrapper;
        private double font_size;
        private int icon_size;
        private Gtk.CssProvider? style_provider;
        private Gtk.Image icon_widget;
        private Gtk.Label label_widget;

        public AppItem (int size, double font_size, int box_width, int box_height) {
            this.get_style_context ().add_class ("app_item");
            this.icon_size = size;
            this.font_size = font_size;

            this.expand = false;
            // Height is also the padding between icon and label's height
            this.set_size_request (box_width, box_height);
            
            this.icon_widget = new Gtk.Image.from_pixbuf (this.icon);
            this.icon_widget.set_halign (Gtk.Align.CENTER);
            this.icon_widget.set_valign (Gtk.Align.CENTER);
            
            this.label_widget = new Gtk.Label ("");
            this.label_widget.expand = false;
            this.label_widget.set_ellipsize (Pango.EllipsizeMode.END);
            this.label_widget.set_alignment (0.5f, 0.5f);
            this.label_widget.selectable = false;
            this.label_widget.can_focus = false;
            this.label_widget.set_single_line_mode (true);
            this.label_widget.margin_bottom = (int) Math.floor(box_height * 0.12);
            this.label_widget.margin_start = 10;
            this.label_widget.margin_end = 10;
            this.label_widget.set_max_width_chars(1);

            // VBox properties
            this.wrapper = new Gtk.Box (Gtk.Orientation.VERTICAL, 0);
            this.wrapper.expand = false;
            this.wrapper.pack_start (this.icon_widget, true, true, 0);
            this.wrapper.pack_end (this.label_widget, false, true, 0);
            this.add (this.wrapper);
        }
        
        public void change_app (Gdk.Pixbuf new_icon, string new_name, string new_tooltip) {
            // Icon
            this.icon = new_icon;
            this.prominent = ZorinAppGridLite.Frontend.Utilities.average_color (this.icon);
            this.icon_widget.set_from_pixbuf (this.icon);
            this.update_background_color();

            // Label
            this.label_widget.set_label (new_name);

            // Tooltip
            this.set_tooltip_text (new_tooltip);
        }
        
        private void update_background_color () {
            double dark = 0.32;
            Gdk.RGBA focus_color = {this.prominent.R * dark, this.prominent.G * dark, this.prominent.B * dark, 0.5};
            Gdk.RGBA active_color = {this.prominent.R * dark, this.prominent.G * dark, this.prominent.B * dark, 0.75};
            
            string new_style = @".app_item:focus,.app_item:hover{background-color:$(focus_color.to_string())}.app_item:active{background-color:$(active_color.to_string())}";
            
            var new_provider = new Gtk.CssProvider ();
            try {
                new_provider.load_from_data (new_style, -1);
            }
            catch (Error e) {
                warning ("Could not create CSS Provider: %s\nStylesheet:\n%s",
                         e.message, new_style);
                return;
            }
            var context = this.get_style_context();
            if (this.style_provider != null) {
                context.remove_provider(this.style_provider);
            }
            this.style_provider = new_provider;
            context.add_provider (this.style_provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION);
        }

    }
    
    public class AppItemContainer : Gtk.Box {
        private AppItem app_item;
        private Gtk.Grid spacer;
        
        public AppItemContainer (AppItem item) {
            this.app_item = item;
            this.spacer = new Gtk.Grid();
            this.pack_start(app_item, true, true, 0);
            this.pack_start(spacer, true, true, 0);
            this.hide_app_item();
        }
        
        public void show_app_item() {
            this.spacer.visible = false;
            this.app_item.visible = true;
        }
        
        public void hide_app_item() {
            this.app_item.visible = false;
            this.spacer.visible = true;
        }
        
        public override void grab_focus() {
            this.app_item.grab_focus();
        }
        
        public void change_app (Gdk.Pixbuf new_icon, string new_name, string new_tooltip) {
            this.app_item.change_app(new_icon, new_name, new_tooltip);
        }

        public bool is_item_visible() {
            return this.app_item.visible;
        }
    }

}
